@@ -1,5 +1,5 @@
 /*!
- * Convenient Formsets 1.2.1 (https://github.com/tiesjan/django-convenient-formsets/)
+ * Convenient Formsets 2.0 (https://github.com/tiesjan/django-convenient-formsets/)
  * Copyright 2021-present Ties Jan Hefting
  * Licensed under BSD 3-Clause License
  */
@@ -32,7 +32,7 @@ const ConvenientFormset = function(options) {
             'defaultValue': undefined,
             'requiredIf': 'canAddForms',
         },
-        'emptyFormSelector': {
+        'emptyFormTemplateSelector': {
             'defaultValue': undefined,
             'requiredIf': 'canAddForms',
         },
@@ -82,7 +82,7 @@ const ConvenientFormset = function(options) {
          * otherwise hidden.
          */
         const visibleForms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
         const maxNumForms = parseInt(
             managementFormElements.maxNumFormsInput.value, 10
@@ -96,11 +96,12 @@ const ConvenientFormset = function(options) {
          * `for` attribute of labels and the `id`/`name` attributes of inputs.
          */
         const prefix = formsetOptions.formsetPrefix;
-        const idRegex = new RegExp(prefix + '-(\\d+|__prefix__)');
+        const idRegex = new RegExp(`${prefix}-(\\d+|__prefix__)`);
 
         let forms = formsetElements.formsContainer.querySelectorAll(
             formsetOptions.formSelector
         );
+        forms = Array.from(forms);
 
         // If forms can be ordered, order the retrieved forms by the `name`
         // attribute of ORDER input elements. This ensures indexes are updated
@@ -128,13 +129,13 @@ const ConvenientFormset = function(options) {
                 }
             };
 
-            forms = Array.prototype.slice.call(forms).sort(compareForms);
+            forms = forms.sort(compareForms);
         }
 
 
         for (let i = 0; i < forms.length; i++) {
             const form = forms[i];
-            const idReplacement = prefix + '-' + i;
+            const idReplacement = `${prefix}-${i}`;
 
             // Update the `for` attribute for all `label` elements
             const labelElements = form.querySelectorAll('label');
@@ -192,7 +193,7 @@ const ConvenientFormset = function(options) {
          * it updates the visibility of the `addFormButton`.
          */
         const visibleForms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
         const maxNumForms = parseInt(
             managementFormElements.maxNumFormsInput.value, 10
@@ -258,6 +259,16 @@ const ConvenientFormset = function(options) {
         // Update the form indexes and the management form
         updateFormIndexes();
         updateManagementForm();
+
+        // Dispatch event
+        formsetElements.formsContainer.dispatchEvent(
+            new CustomEvent('convenient_formset:added', {
+                bubbles: true,
+                detail: {
+                    formsetPrefix: formsetOptions.formsetPrefix,
+                },
+            })
+        );
     }
 
     function deleteFormButtonClicked(form) {
@@ -287,6 +298,15 @@ const ConvenientFormset = function(options) {
         // Update the form indexes and the management form
         updateFormIndexes();
         updateManagementForm();
+
+        // Dispatch event
+        document.dispatchEvent(
+            new CustomEvent('convenient_formset:removed', {
+                detail: {
+                    formsetPrefix: formsetOptions.formsetPrefix,
+                },
+            })
+        );
     }
 
     function moveFormDownButtonClicked(form) {
@@ -296,9 +316,9 @@ const ConvenientFormset = function(options) {
          * values.
          */
         let visibleForms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
-        visibleForms = Array.prototype.slice.call(visibleForms);
+        visibleForms = Array.from(visibleForms);
         const formIndex = visibleForms.indexOf(form);
 
         if (formIndex < visibleForms.length - 1) {
@@ -314,6 +334,16 @@ const ConvenientFormset = function(options) {
             const nextFormOrderValue = nextFormOrderElement.value;
             formOrderElement.value = nextFormOrderValue;
             nextFormOrderElement.value = formOrderValue;
+
+            // Dispatch event
+            form.dispatchEvent(
+                new CustomEvent('convenient_formset:movedDown', {
+                    bubbles: true,
+                    detail: {
+                        formsetPrefix: formsetOptions.formsetPrefix,
+                    },
+                })
+            );
         }
     }
 
@@ -324,9 +354,9 @@ const ConvenientFormset = function(options) {
          * values.
          */
         let visibleForms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
-        visibleForms = Array.prototype.slice.call(visibleForms);
+        visibleForms = Array.from(visibleForms);
         const formIndex = visibleForms.indexOf(form);
 
         if (formIndex > 0) {
@@ -342,6 +372,16 @@ const ConvenientFormset = function(options) {
             const previousFormOrderValue = previousFormOrderElement.value;
             formOrderElement.value = previousFormOrderValue;
             previousFormOrderElement.value = formOrderValue;
+
+            // Dispatch event
+            form.dispatchEvent(
+                new CustomEvent('convenient_formset:movedUp', {
+                    bubbles: true,
+                    detail: {
+                        formsetPrefix: formsetOptions.formsetPrefix,
+                    },
+                })
+            );
         }
     }
 
@@ -354,7 +394,7 @@ const ConvenientFormset = function(options) {
          */
         const missingOptions = [];
 
-        for (let optionKey in availableOptions) {  // eslint-disable-line prefer-const -- using `let` for IE11
+        for (const optionKey in availableOptions) {
             // Attempt to retrieve option values from `customOptions`, falling
             // back to the default value
             let optionValue;
@@ -385,20 +425,20 @@ const ConvenientFormset = function(options) {
         // Throw error if there are missing required options
         if (missingOptions.length) {
             const formattedMissingOptions = missingOptions.map(
-                function(item) { return '`' + item + '`'; }
-            ).join(', ');
+                function(item) { return `"${item}"`; }
+            );
             const message = (
-                'Missing required options: ' + formattedMissingOptions
+                `Missing required options: ${formattedMissingOptions.join(', ')}`
             );
             throw new Error(message);
         }
     }
 
-    function checkEmptyFormElements() {
+    function checkEmptyFormTemplateElements() {
         /*
-         * Asserts that the empty form has an `deleteFormButton` if forms can
-         * be deleted, and a `moveFormDownButton` and a `moveFormUpButton` if
-         * forms can be ordered.
+         * Asserts that the empty form template has an `deleteFormButton` if
+         * forms can be deleted, a `moveFormDownButton` and a
+         * `moveFormUpButton` if forms can be ordered.
          */
         const missingElements = [];
         let element;
@@ -435,11 +475,11 @@ const ConvenientFormset = function(options) {
         // Throw error if DOM elements are missing
         if (missingElements.length) {
             const formattedMissingElements = missingElements.map(
-                function(item) { return '`' + item + '`'; }
-            ).join(', ');
+                function(item) { return `"${item}"`; }
+            );
             const message = (
-                'Unable to find DOM elements in empty form with selectors: ' +
-                formattedMissingElements
+                'Unable to find DOM elements in empty form template with ' +
+                `selectors: ${formattedMissingElements.join(', ')}`
             );
             throw new Error(message);
         }
@@ -456,7 +496,7 @@ const ConvenientFormset = function(options) {
         let selector;
 
         const forms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
         for (let i = 0; i < forms.length; i++) {
             const form = forms[i];
@@ -493,11 +533,11 @@ const ConvenientFormset = function(options) {
         // Throw error if DOM elements are missing
         if (missingElements.length) {
             const formattedMissingElements = missingElements.map(
-                function(item) { return '`' + item + '`'; }
-            ).join(', ');
+                function(item) { return `"${item}"`; }
+            );
             const message = (
                 'Unable to find DOM elements in one or more visible forms ' +
-                'with selectors: ' + formattedMissingElements
+                `with selectors: ${formattedMissingElements.join(', ')}`
             );
             throw new Error(message);
         }
@@ -508,9 +548,12 @@ const ConvenientFormset = function(options) {
          * Initializes `formsetElements` by selecting DOM elements from the
          * formset using the specified selectors.
          *
-         * Missing DOM elements will throw an error.
+         * Missing DOM elements and malformed empty form template will throw an
+         * error.
          */
+        let emptyFormChildElementCount = -1;
         const missingElements = [];
+
         let selector;
 
         // Select DOM elements
@@ -521,11 +564,19 @@ const ConvenientFormset = function(options) {
         }
 
         if (formsetOptions.canAddForms) {
-            selector = formsetOptions.emptyFormSelector;
-            formsetElements.emptyForm = document.querySelector(selector);
-            if (formsetElements.emptyForm === null) {
+            selector = formsetOptions.emptyFormTemplateSelector;
+            const emptyFormTemplate = document.querySelector(selector);
+            if (emptyFormTemplate === null) {
                 missingElements.push(selector);
             }
+            else {
+                emptyFormChildElementCount = emptyFormTemplate.content.childElementCount;
+                if (emptyFormChildElementCount === 1) {
+                    formsetElements.emptyForm = (
+                        emptyFormTemplate.content.firstElementChild
+                    );
+                }
+            }
 
             selector = formsetOptions.addFormButtonSelector;
             formsetElements.addFormButton = document.querySelector(selector);
@@ -537,11 +588,21 @@ const ConvenientFormset = function(options) {
         // Throw error if DOM elements are missing
         if (missingElements.length) {
             const formattedMissingElements = missingElements.map(
-                function(item) { return '`' + item + '`'; }
-            ).join(', ');
+                function(item) { return `"${item}"`; }
+            );
             const message = (
                 'Unable to find DOM elements with selectors: ' +
-                formattedMissingElements
+                `${formattedMissingElements.join(', ')}`
+            );
+            throw new Error(message);
+        }
+
+        // Throw error if the empty form template is malformed
+        else if (formsetOptions.canAddForms && emptyFormChildElementCount !== 1) {
+            const message = (
+                'Expected 1 element inside ' +
+                `"${formsetOptions.emptyFormTemplateSelector}", ` +
+                `found: ${emptyFormChildElementCount}`
             );
             throw new Error(message);
         }
@@ -558,25 +619,25 @@ const ConvenientFormset = function(options) {
         let selector;
 
         // Select DOM elements
-        selector = 'input[name="' + formsetOptions.formsetPrefix + '-TOTAL_FORMS"]';
+        selector = `input[name="${formsetOptions.formsetPrefix}-TOTAL_FORMS"]`;
         managementFormElements.totalFormsInput = document.querySelector(selector);
         if (managementFormElements.totalFormsInput === null) {
             configurationError = true;
         }
 
-        selector = 'input[name="' + formsetOptions.formsetPrefix + '-INITIAL_FORMS"]';
+        selector = `input[name="${formsetOptions.formsetPrefix}-INITIAL_FORMS"]`;
         managementFormElements.initialFormsInput = document.querySelector(selector);
         if (managementFormElements.initialFormsInput === null) {
             configurationError = true;
         }
 
-        selector = 'input[name="' + formsetOptions.formsetPrefix + '-MIN_NUM_FORMS"]';
+        selector = `input[name="${formsetOptions.formsetPrefix}-MIN_NUM_FORMS"]`;
         managementFormElements.minNumFormsInput = document.querySelector(selector);
         if (managementFormElements.minNumFormsInput === null) {
             configurationError = true;
         }
 
-        selector = 'input[name="' + formsetOptions.formsetPrefix + '-MAX_NUM_FORMS"]';
+        selector = `input[name="${formsetOptions.formsetPrefix}-MAX_NUM_FORMS"]`;
         managementFormElements.maxNumFormsInput = document.querySelector(selector);
         if (managementFormElements.maxNumFormsInput === null) {
             configurationError = true;
@@ -585,9 +646,9 @@ const ConvenientFormset = function(options) {
         // Throw error if DOM elements are missing
         if (configurationError) {
             const message = (
-                'Management form for formset with prefix `' +
-                formsetOptions.formsetPrefix +
-                '` missing or has been tampered with.'
+                'Management form for formset with prefix "' +
+                `${formsetOptions.formsetPrefix}` +
+                '" missing or has been tampered with.'
             );
             throw new Error(message);
         }
@@ -622,7 +683,7 @@ const ConvenientFormset = function(options) {
         }
 
         const forms = formsetElements.formsContainer.querySelectorAll(
-            formsetOptions.formSelector + ':not([hidden])'
+            `${formsetOptions.formSelector}:not([hidden])`
         );
         for (let i = 0; i < forms.length; i++) {
             const form = forms[i];
@@ -657,29 +718,29 @@ const ConvenientFormset = function(options) {
             initializeFormsetOptions(customOptions);
         }
         catch (error) {
-            throw Error('[ConvenientFormset] ' + error.message);
+            throw Error(`[ConvenientFormset] ${error.message}`);
         }
 
         try {
             selectFormsetElements();
         }
         catch (error) {
-            throw Error('[ConvenientFormset] ' + error.message);
+            throw Error(`[ConvenientFormset] ${error.message}`);
         }
 
         try {
             selectManagementFormElements();
         }
         catch (error) {
-            throw Error('[ConvenientFormset] ' + error.message);
+            throw Error(`[ConvenientFormset] ${error.message}`);
         }
 
         if (formsetOptions.canAddForms) {
             try {
-                checkEmptyFormElements();
+                checkEmptyFormTemplateElements();
             }
             catch (error) {
-                throw Error('[ConvenientFormset] ' + error.message);
+                throw Error(`[ConvenientFormset] ${error.message}`);
             }
         }
 
@@ -687,7 +748,7 @@ const ConvenientFormset = function(options) {
             checkVisibleFormsElements();
         }
         catch (error) {
-            throw Error('[ConvenientFormset] ' + error.message);
+            throw Error(`[ConvenientFormset] ${error.message}`);
         }
 
         if (formsetOptions.canDeleteForms) {
