@@ -2,7 +2,6 @@ from django.urls import path
 
 from .views import TestView
 
-
 urlpatterns = [
-    path('', TestView.as_view()),
+    path("", TestView.as_view()),
 ]
